/* -*- Mode: C; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*-
 *
 * The contents of this file are subject to the Mozilla Public 
 * License version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or 
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 *
 * The Initial Developer of the Original Code is R.V.Guha. Portions
 * created by R.V.Guha are copyright (c) 2000 R.V.Guha. All Rights
 * Reserved.
 *
 * Contributors : Jeen Broekstra
 */


#include "dbglue.h"
#include <db.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <dirent.h>
#include "../utils/hash.h"
#include "rdf.h"
#include "../utils/utils.h"
#include <pthread.h>
#include <dirent.h>

typedef struct _IndexStruct* Index;
#include "indexPool.h"
#define MAX_ANS_SIZE 4096

#define ALL_PERMS (S_IRWXU | S_IRWXO | S_IRWXG)

typedef struct _DB_CursorStruct {
  DBT key;
  DBT data;
  DBC *cursor;
  DB  *db;
  int inited;
  void* value;
  size_t valueSize;
  int type;
  int count;
  int qtype;
  char* ind;
  struct _DB_CursorStruct *next;
  char ans[MAX_ANS_SIZE];
} DB_CursorStruct;

typedef struct _BDB_FileStruct {
  char* url;
  char* name;
  struct _BDB_FileStruct* next;
} BDB_FileStruct;

typedef BDB_FileStruct* BDB_File;

typedef struct _BDB_Struct {
  char* name; 
  int    nextAssertionRecord;
  DB*    findex;
  DB*    rindex;
  DB*    arcindex;
  BDB_File files;
} BDB_struct;

typedef BDB_struct* BDB;

HashTable gDBHash = 0;

char* gDatabaseDirectory = 0;


static DB_Cursor gDBCursorPool = 0;
pthread_mutex_t gDBCursorPoolMutex = PTHREAD_MUTEX_INITIALIZER;
DB_Cursor
getNewDBCursor () {
  DB_Cursor existing;
  pthread_mutex_lock(&gDBCursorPoolMutex);
  existing = gDBCursorPool;
  if (existing) {
    gDBCursorPool = existing->next;
    bzero(existing, sizeof(DB_CursorStruct));
  } 
  pthread_mutex_unlock(&gDBCursorPoolMutex);
  if (!existing) {
    existing = (DB_Cursor) getMem(sizeof(DB_CursorStruct));
  }
  return existing;
}

void 
freeDBCursor (DB_Cursor c) {
  pthread_mutex_lock(&gDBCursorPoolMutex);
  c->next = gDBCursorPool;
  gDBCursorPool = c;
  pthread_mutex_unlock(&gDBCursorPoolMutex);
}



void 
DB_Initialize() {
    if (!gDBHash) gDBHash = NewHashTable(256);
    if (!gDatabaseDirectory) {
      gDatabaseDirectory =   getenv("RDFDB_DIR");
      if (!gDatabaseDirectory) {
      printf("\n********* Please set the environment variable RDFDB_DIR to the directory where you want the data to be stored!**********\ne.g., export RDFDB_DIR=/var/rdfdb/\n");
      exit(1);
      } 
    }
}

void
CloseAllDatabases () {
  int n = 0;
  while (n < gDBHash->size) {
    HashEntry b = gDBHash->buckets[n];
    while (b) {
      BDB bdb = b->value;
      bdb->findex->close(bdb->findex, 0);
      bdb->rindex->close(bdb->rindex, 0);
      bdb->arcindex->close(bdb->rindex, 0);
      b = b->next;
    }
    n++;
  }
}


int DB_DatabaseExistsp (const char* db) {
  char fullName[256];
  DIR *dir = 0;
  snprintf(fullName, 256, "%s%s", gDatabaseDirectory, db);
  dir = opendir(fullName);
  if (!dir) return 0;
  closedir(dir);
  return 1;
}

DB_ENV* gdb_env = 0;

/*
void
init_db_env () {
  db_env_create(&gdb_env, 0);
  gdb_env->open(gdb_env, gDatabaseDirectory, 0, 0,0);
  } */

int
openTable (const char* dbName, DB** db, const char* name, int type,  int dups) {
  int err = db_create(db, 0, 0);

  if (err) {
    printf("Problems creating %s : %s\n", name, db_strerror(err));
    return err;
  }
  
  if (dups) {
    err = (*db)->set_flags(*db, DB_DUP);
    if (err) {
      printf("Problems setting duplicate flag for %s : %s\n", name, db_strerror(err));
      return err;
    }
  }

  err = (*db)->open(*db, name, NULL, type, DB_CREATE | DB_THREAD , 0664);
  if (err) {
    printf("Problems opening %s : %s\n", name, db_strerror(err));
    return err;
  }
  return 0;
}

void
initBDBFiles (BDB db) {
  char fullName[256];
  char *buff = getMem(1024);
  DIR *d;
  struct dirent *f;
  snprintf(fullName, 256, "%s%s", gDatabaseDirectory, db->name);
  d = opendir(fullName);
  if (!d) {
    freeMem(buff);
    return;
  }
  while (0 != (f = readdir(d))) {
    char* name = f->d_name;
    char* suffix = strchr(name, '.');
    if (suffix && streq(suffix+1, "dbfl")) {
      BDB_File dbf = (BDB_File) getMem(sizeof(BDB_FileStruct));
      int fi = open(name, O_RDONLY);
      char *end;
      dbf->name = getMem(suffix-name+1);
      memcpy(dbf->name, name, suffix-name-1);
      dbf->next = db->files;
      db->files = dbf;
      bzero(buff, 1024);
      read(fi, buff, 1024);
      end = strchr(buff, '\n');
      if (end) *end = 0;
      dbf->url = strdup(buff);
      close(fi);
    }
  }
  closedir(d);
  freeMem(buff);
}

int
openDB (const char* name, int createp) {
  char fullName[256];
  DIR *dir = 0;
  BDB db  = (BDB)getMem(sizeof(BDB_struct));
  int err =0;

  snprintf(fullName, 256, "%s%s", gDatabaseDirectory, name);
  if (chdir(fullName)) {
    if (createp) {
      if (0 != (err = mkdir(fullName, ALL_PERMS))) {
        printf("Cannot mkdir  %s\n", fullName);
        return RDF_WRONG_FILE_PERMISSIONS_ERROR;
    }
      if (0 != (err = chdir(fullName))) {
        printf("Cannot chdir to %s\n", fullName);
        return RDF_WRONG_FILE_PERMISSIONS_ERROR;
      }
    } else {
      printf("%s does not exist!\n", fullName);
      return -1;
    }
  }
  
  db->name = strdup(name);
  if (0 != (err = openTable(name, &(db->findex), "findex.db", DB_BTREE, 1))) {
    printf("Cannot open findex.db\n");
    return err;
  }
  if (0 != (err = openTable(name, &(db->rindex), "rindex.db", DB_BTREE, 1))) {
    printf("Cannot open rindex\n");
    return err;
  }
  if (0 != (err = openTable(name, &(db->arcindex), "arcindex.db", DB_BTREE, 1))) {
    printf("Cannot open arcindex \n");
    return err;
  }

  initBDBFiles(db);
  HashAdd(gDBHash, name, db);
  return 0;
}



int 
DB_CreateDatabase (const char* db) {
  if (!HashLookup(gDBHash, db)) {
    openDB(db, 1);
  } else return 0;
}

pthread_mutex_t gOpenDBMutex = PTHREAD_MUTEX_INITIALIZER;


BDB
db2bdb (const char* name, int createp) {
  BDB bdb;
  pthread_mutex_lock(&gOpenDBMutex);
  if (NULL != (bdb = HashLookup(gDBHash, name))) {
    pthread_mutex_unlock(&gOpenDBMutex);
    return bdb;
  } else {
    openDB(name, createp);
    pthread_mutex_unlock(&gOpenDBMutex);
    return HashLookup(gDBHash, name);
  }
}

int
DB_GetFileCode (const char* db, const char *url, char **code, FILE **file) {
  BDB bdb = db2bdb(db, 0);
  BDB_File bf ;
  int n = 1;
  char fileName[12];
  char fullName[256];
  char lbuff[512];

  if (!bdb) return 0;
  bf = bdb->files;
  pthread_mutex_lock(&gOpenDBMutex);
  while (bf) {
    if (streq(bf->url, url)) {
      pthread_mutex_unlock(&gOpenDBMutex);
      snprintf(fullName, 256, "%s%s/%i.dbfl", gDatabaseDirectory, bdb->name, n);
      *file = fopen(fullName, "r+");
      if (*file) {
        fgets(lbuff, 512, *file);
      }
      *code = bf->name;
      return 0;
    }
    bf = bf->next;
    n++;
  }

  bf = (BDB_File)getMem(sizeof(BDB_FileStruct));
  bf->next = bdb->files;
  bdb->files = bf;
  
  bf->url = strdup(url);
  bf->name = getMem(12);
  snprintf(bf->name, 12, "%i", n);

  snprintf(fullName, 256, "%s%s/%i.dbfl", gDatabaseDirectory, bdb->name, n);
  *file = fopen(fullName, "a+");
  chmod(fullName, ALL_PERMS);

  if (*file) fprintf(*file, "%s\n", url);

  *code = bf->name;
   
  pthread_mutex_unlock(&gOpenDBMutex);
  return 0;
}
  
  
int 
DB_CloseDatabase (const char* dbName) {
  BDB bdb = db2bdb(dbName, 0);
  bdb->findex->close(bdb->findex, 0);
  bdb->rindex->close(bdb->rindex, 0);
  bdb->arcindex->close(bdb->rindex, 0);
}

size_t itemLen (void* item, RDF_ValueType type) {
  if ((type == RDF_RESOURCE_TYPE) || (type == RDF_STRING_TYPE))  {
    return strlen((char*) item);
  } else return 12;
}

int
getIndex (DB* db, const char* index, DB_Cursor *c, int qtype) {
  DBC *dbc;
  DB_Cursor xdbc = getNewDBCursor();
  int err = 0;

  err = db->cursor(db, 0, &dbc, 0);
  if (err) {
    freeMem(xdbc);
    return err;
  }

  xdbc->qtype = qtype;
  xdbc->key.size  = strlen(index);
  if (qtype == 2) {
    xdbc->key.data = xdbc->ans;
    xdbc->ind = index;
    memcpy(xdbc->key.data, index, strlen(index));
    xdbc->key.flags = DB_DBT_USERMEM;
    xdbc->key.ulen = MAX_ANS_SIZE;
  } else {
    xdbc->key.data = index;
    xdbc->data.data = xdbc->ans;
    xdbc->data.flags = DB_DBT_USERMEM;
    xdbc->data.ulen = MAX_ANS_SIZE;
  }
  xdbc->cursor = dbc;
  xdbc->db = db;
  *c = xdbc;
  return 0;
}

int asc = 0;
int nasc = 0;

int
addremoveIndex (DB* db, const char* index, const char* value, int op, int dupCheck) {
  DBC *dbc;
  DBT key;
  DBT data;
  int err = 0;
  int found = 0;
  int first = 1;
  int somethingFound = 0;
  extern int gVerbose;
  bzero(&key, sizeof(DBT));
  bzero(&data, sizeof(DBT));

  db->cursor(db, 0, &dbc, 0);

  key.data = index;
  key.size  = strlen(index);

  data.data = value;
  data.size  = strlen(value);

  /*  if (dupCheck)   printf("\n***********************\n"); */

  if (op || dupCheck) {
  for (;;) {
    int err = dbc->c_get(dbc, &key, &data, (first ? DB_SET : DB_NEXT_DUP));
    first = 0;
    if (err) {
      if (err == DB_NOTFOUND) {
        break;
      } else {
        printf("Problems getting cursor during %s for key=%s, value=%s: %s\n", 
               (op ? "removal" : "addition"), index, value, db_strerror(err));
        return err;
      }
    }
    somethingFound = 1;
    /*    printf("%s ===> %s\n", (char*)key.data, &((char*)data.data)[4]); */
    if (strncmp((char*)(data.data), value, data.size) == 0) {
      if (op == 1) {
        err = dbc->c_del(dbc, 0);
        if (err) {
          printf("Problems removing key=%s, value=%s: %s\n", 
                 index, value, db_strerror(err));
          dbc->c_close(dbc);
          return err;
        }
      } else {
        found = 1;
        if (gVerbose > 2) {
          printf("\tAlready in database\n");
        }
      }
      break;
    }
  }
  }
  if ((op == 0) && (!found)) {

    bzero(&data, sizeof(DBT));
    bzero(&data, sizeof(DBT));
    key.data = index;
    key.size  = strlen(index);
    data.data = value;
    data.size  = strlen(value);
    err = dbc->c_put(dbc, &key, &data, DB_KEYFIRST);    
    asc++;
    if (asc == 1000) {
      nasc++;
      printf("[%i]", nasc);
      asc = 0;
      fflush(stdout);
    }
    if (err) {
      printf("Problems putting  for key=%s, value=%s: %s\n", 
              index, value, db_strerror(err));
      dbc->c_close(dbc);
      return err;
    } else if (gVerbose > 2) {
            printf("\nAdded %s (key %s)\n", (char*)data.data, (char*)key.data);
    }
  }
  
  dbc->c_close(dbc);

  if (op == 0 && dupCheck) {
    if (found) {
      return 2;
    } else if (somethingFound) {
      return 1;
    }
  }
  return 0;
}

int
DB_Flush (const char* db) {
  BDB bdb = db2bdb(db, 0);
  if (bdb) {
    bdb->findex->sync(bdb->findex, 0);
    bdb->rindex->sync(bdb->rindex, 0);
    bdb->arcindex->sync(bdb->arcindex, 0);
    return 0;
  } else return RDF_DB_DOES_NOT_EXIST;
}
  
int 
DB_AddForwardIndex(const char* dbName, const char* forwardIndex, const char* value, 
                   int dupCheck) {
  BDB db = db2bdb(dbName, 0);
  if (!db) return RDF_DB_DOES_NOT_EXIST;
  return addremoveIndex(db->findex, forwardIndex, value, 0, dupCheck);
}

int DB_RemoveForwardIndex(const char* dbName, const char* forwardIndex, const char* value) {
  BDB db = db2bdb(dbName, 0);
  if (!db) return RDF_DB_DOES_NOT_EXIST;
  return addremoveIndex(db->findex, forwardIndex, value, 1, 1);
}

int DB_AddReverseIndex(const char* dbName, const char* reverseIndex, const char* value, int dupCheck) {
  BDB db = db2bdb(dbName, 0);
  if (!db) return RDF_DB_DOES_NOT_EXIST;
  return addremoveIndex(db->rindex, reverseIndex, value, 0, dupCheck);
}

int DB_RemoveReverseIndex(const char* dbName, const char* reverseIndex, const char* value) {
  BDB db = db2bdb(dbName, 0);
  if (!db) return RDF_DB_DOES_NOT_EXIST;
  return addremoveIndex(db->rindex, reverseIndex, value, 1, 1);
}


int DB_AddArcIndex(const char* dbName, const char* arcIndex, const char* value, 
                   int dupCheck) {
  BDB db = db2bdb(dbName, 0);
  if (!db) return RDF_DB_DOES_NOT_EXIST;
  return addremoveIndex(db->arcindex, arcIndex, value, 0, dupCheck);
}

int DB_RemoveArcIndex(const char* dbName, const char* reverseIndex, const char* value) {
  BDB db = db2bdb(dbName, 0);
  if (!db) return RDF_DB_DOES_NOT_EXIST;
  return addremoveIndex(db->arcindex, reverseIndex, value, 1, 1);
}


int
DB_GetForwardCursor(const char* dbName, const char* forwardIndex, DB_Cursor *c) {
  BDB db = db2bdb(dbName, 0);
  if (!db) return RDF_DB_DOES_NOT_EXIST;
  return getIndex(db->findex, forwardIndex, c, 0);
}

int
DB_GetReverseCursor(const char* dbName, const char* reverseIndex, DB_Cursor *c) {
  BDB db = db2bdb(dbName, 0);
  if (!db) return RDF_DB_DOES_NOT_EXIST;
  return getIndex(db->rindex, reverseIndex, c, 0);
}

int 
DB_GetArcsOutCursor(const char* dbName, const char* index, DB_Cursor *c) {
 BDB db = db2bdb(dbName, 0);
 if (!db) return RDF_DB_DOES_NOT_EXIST;
 return getIndex(db->findex, index, c, 2);
}

int 
DB_GetArcsInCursor(const char* dbName, const char* index, DB_Cursor *c) {
 BDB db = db2bdb(dbName, 0);
 if (!db) return RDF_DB_DOES_NOT_EXIST;
 return getIndex(db->rindex, index, c, 2);
}
 
int
DB_GetArcCursor(const char* dbName, const char* arcIndex, DB_Cursor *c) {
  BDB db = db2bdb(dbName, 0);
  if (!db) return RDF_DB_DOES_NOT_EXIST;
  return getIndex(db->arcindex, arcIndex, c, 0);
}

int
DB_CursorAdvance(DB_Cursor c) {
  int err;
  int flags;

  if (!c->qtype) {
    if (!c->inited) 
      flags = DB_SET;
    else
      flags = DB_NEXT_DUP;
  } else {
    if (!c->inited) 
      flags = DB_SET_RANGE;
    else
      flags = DB_NEXT;
  }    

  c->data.size = 0;
  err = c->cursor->c_get(c->cursor, &(c->key), &(c->data), flags);
  c->inited = 1;
  if (err == DB_NOTFOUND) {
    return RDF_NO_MORE_VALUES;
  } else if (err != 0) {
    printf("Problem with query %s\n", db_strerror(err));
    return err;
  } else if (c->qtype == 2) {
    char* value = c->key.data;
    size_t len  = c->key.size - strlen(c->ind)+1;
    if (!startsWith(value, c->ind)) return RDF_NO_MORE_VALUES;   
    c->type = RDF_RESOURCE_TYPE;     
    value[c->key.size] = 0; 
    if ((!startsWith(value, c->ind)) || (c->value && (streq(c->value, &value[strlen(c->ind)])))) {
      flags = DB_NEXT;
      for (;;) {
        c->data.size = 0;
        err = c->cursor->c_get(c->cursor, &(c->key), &(c->data), flags);
        if (err == DB_NOTFOUND) return RDF_NO_MORE_VALUES;
        if (err) return err;
        value = c->key.data;
        if (!startsWith(value, c->ind)) return RDF_NO_MORE_VALUES;   
        value[c->key.size] = 0;
        len  = c->key.size - strlen(c->ind)+1;
        if (!streq(c->value, &value[strlen(c->ind)])) break;
      }
    }
    c->valueSize = len;
    c->count++;
    if (c->value) freeMem(c->value); 
    c->value = getMem(strlen(value));
    memcpy(c->value, &value[strlen(c->ind)], len);
    /*    printf(">>>%s<<<%s>>>%s\n", (char*)c->value, value, c->ind); */
    return 0;
  } else {
    char* value = c->data.data;
    size_t len  = c->data.size;
    RDF_ValueType type = (RDF_ValueType) atoi(value);
    c->type = type;      
    c->count++;
    if ((type == RDF_RESOURCE_TYPE) || (type == RDF_STRING_TYPE)) {
      c->value = strchr(&value[2], '\n')+1;
      c->valueSize = len - ((char*)c->value - value);
      return 0;
    } else if (type == RDF_INT_TYPE) {
      c->value = (void*)atoi(strchr(&value[2], '\n')+1);
      c->valueSize = sizeof(int);
      return 0;
    }
  }
}


int 
DB_CursorCurrentValue (DB_Cursor c, void** value) {
  if (!c) {
    return RDF_ERROR;
  } else {
    *value = c->value;
    return 0;
  }
}

int 
DB_CursorCurrentValueType (DB_Cursor c, int *type) {
  if (!c) {
    return RDF_ERROR;
  } else {
    *type = c->type;
    return 0;
  }
}

int 
DB_CursorCurrentValueSize (DB_Cursor c, int *sz) {
  if (!c) {
    return RDF_ERROR;
  } else {
    *sz = c->valueSize;
    return 0;
  }
}

int DB_ReleaseCursor(DB_Cursor c) {
  /*   char* keydata = c->key.data; */  
  c->cursor->c_close(c->cursor); 
  if (c->qtype == 2 ) {
    if (c->value) freeMem(c->value);
  }
  freeDBCursor(c);  
}








