/* -*- Mode: C; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*-
 *
 * The contents of this file are subject to the Mozilla Public 
 * License version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or 
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 *
 * The Initial Developer of the Original Code is R.V.Guha. Portions
 * created by R.V.Guha are copyright (c) 2000 R.V.Guha. All Rights
 * Reserved.
 */

#include <string.h>
#include <stdio.h>
#include "rdf-int.h"
#include "xmlparse.h"
#include "contentSource.h"
#include "../utils/utils.h"
#include "../utils/hash.h"
#include "parsers.h"

RDF_Parser NewRDFParser (RDF_SerializationType type) ;

RDF_ValueType
termType (const char *term) {
  int l = strlen(term);
  int c = 0;
  int isn = 1;

  if (term[0] == '\'') return RDF_STRING_TYPE;
  
  while (isn && c < l) {
    if (!isdigit(term[c++])) isn = 0;
  }
  
  if (isn) return RDF_INT_TYPE;
  return RDF_RESOURCE_TYPE;
}
      
void*
parseTermOfType (char* term, RDF_ValueType type) {
  if (type == RDF_RESOURCE_TYPE) {
    return strdup(term);
  } else if (type == RDF_INT_TYPE) {
    return (void*)atoi(term);
  } else if (type == RDF_STRING_TYPE) {
    term[strlen(term)-1] = 0;
    return strdup(&term[1]);    
  }
}

RDF_Resource 
newAnonResource (RDF_Parser parser, const char* name) {
  size_t len ;
  char* buff ;
  char* post = strchr(name, '#');
  if (!post) post = name;
  len = strlen(parser->url) + strlen(post) + 10;
  buff = getMem(len);
  snprintf(buff, len, "A:%s-%s-%i", parser->url, post, parser->count++);
  return RDF_GetResource(buff);
}


void 
RSS_StartElementHandler (void *userData, const XML_Char *name, const XML_Char **atts) {
  RDF_Parser rdfp = (RDF_Parser) userData;
  if (streq(name, "channel")) {
    RDF_Resource nres ;
    nres = newAnonResource(rdfp, name);
    RDF_LoadFileAssert(rdfp, rdfp->db, nres, RDF_GetResource("type"), RDF_GetResource("Channel"), RDF_RESOURCE_TYPE);
    rdfp->stack[rdfp->depth++] = nres;
    rdfp->status = 1;
  } else if (streq(name, "title") || streq(name, "description") || streq(name, "link")) {
    rdfp->stack[rdfp->depth++] = RDF_GetResource(name);
    rdfp->status = 2;
  } else if (streq(name, "item")) {
    RDF_Resource nres ;
    nres = newAnonResource(rdfp, name);
    RDF_LoadFileAssert(rdfp, rdfp->db, nres, RDF_GetResource("type"), RDF_GetResource("Item"), RDF_RESOURCE_TYPE);
    RDF_LoadFileAssert(rdfp, rdfp->db, rdfp->stack[rdfp->depth-1], RDF_GetResource("item"), nres, RDF_RESOURCE_TYPE);
    rdfp->stack[rdfp->depth++] = nres;
    rdfp->status = 1;
  } else {
    rdfp->depth++;
  }
}

void 
RSS_EndElementHandler (void *userData, const XML_Char *name) {
  RDF_Parser rdfp = (RDF_Parser) userData;
  if (rdfp->status == 1) {
    freeResource(rdfp->stack[rdfp->depth]);
  }
  rdfp->depth--;
}

void 
RSS_CharacterDataHandler(void *userData, const XML_Char *s, int len) {
  RDF_Parser rdfp = (RDF_Parser) userData;
  char* buff;
  if (nullString(s, len)) return;

  buff = getMem(len+1);
  bzero(buff, len+1);
  memcpy(buff, s, len);

  if (rdfp->status == 2) {
    RDF_LoadFileAssert(rdfp, rdfp->db, rdfp->stack[rdfp->depth-2], rdfp->stack[rdfp->depth-1], 
               buff, RDF_STRING_TYPE);
    freeMem(buff);
  }
}

FILE *
rdfpGetFile(RDF_Parser rdfp) {
  return rdfp->file;
}


void
unloadExistingAssertions(FILE *f, const char* db) {
  if (f) {
    long pos = ftell(f);
    char *buff = getMem(4096);
    char *line;
    while (0 != (line = fgets(buff, 4096, f))) {
      char *arc, *source, *target;
      arc = line;
      source = strchr(arc, '\t');
      if (source) {
        *source = 0;
        source++;
        target = strchr(source, '\t');
        if (target) {
          RDF_ValueType type ;
          void* value;
          *target = 0;
          target++;
          target[strlen(target)-1] = 0;
          type = termType(target) ;
          value = parseTermOfType (target,  type) ;
          RDF_Unassert(db, source, arc, value, type);
        }
      }
    }
    freeMem(buff);
    fseek(f, pos, SEEK_SET);
  }
}
 
RDF_Result         
RDF_UnLoadURLFromDB (const char* db, const char* url) {
  char *code;
  FILE *f = 0;
  DB_GetFileCode(db, url, &code, &f);
  unloadExistingAssertions(f, db);
  return 0;
}

#define PARSE_BUFFER_SIZE (4096*1024) /* 4 megabytes */
#define CLEANUP_CONNECTION    {close(rdfp->file); XML_Parse(xmlp, buffer, 0, 1);CS_Close(cs); freeMem((void*)rdfp); freeMem(buffer);  XML_ParserFree(xmlp); freeMem(nurl);}

RDF_Result 
RDF_LoadURLIntoDB (RDF_DB db, const char* url, RDF_SerializationType type, 
                   int reloadp, char** code) {
  if (db2bdb(db, 0) == 0) {
    return RDF_DB_DOES_NOT_EXIST;
  } else {
    char *buffer = getMem(PARSE_BUFFER_SIZE+1);
    size_t len = 0;
    char* nurl = strdup(url);
    int err;
    int firstSegment = 1;
    ContentSource cs = CS_NewContentSource(nurl);
    RDF_Parser rdfp = NewRDFParser(type);
    XML_Parser xmlp = XML_ParserCreateNS(0, '#');
    
    
    rdfp->db  = db;
    rdfp->url = nurl;
    rdfp->xml_parser = xmlp;
    DB_GetFileCode(rdfp->db, rdfp->url, &rdfp->code, &rdfp->file);
    /*  unloadExistingAssertions(rdfp->file, rdfp->db); */
    XML_SetUserData(xmlp, (void*)rdfp);
    XML_SetElementHandler(xmlp, rdfp->startElementHandler, rdfp->endElementHandler);
    XML_SetCharacterDataHandler(xmlp, rdfp->characterDataHandler);
    
    if (!cs) {
      CLEANUP_CONNECTION
        return RDF_ERROR;
    }
    
    CS_Start(cs);
    
    for (;;) {
      CS_Status status;
      char *sbuff;
      bzero(buffer, PARSE_BUFFER_SIZE);
      status = CS_GetContent(cs, buffer, PARSE_BUFFER_SIZE, &len);
      if (status == CS_DONE) {
        break;
      } else if (status == CS_ERROR) {
        CLEANUP_CONNECTION
          printf("Quitting fetching %s\n", url);
        return RDF_ERROR;
      } else if (len == 0) {
        sleep(1);
      } else {
        extern int gVerbose;
        if (gVerbose) {
          /*        printf("%s\n", buffer); */
        }
        if (firstSegment) {
          char* bu = strstr(buffer, "<?xml");
          if (bu) sbuff = bu; else sbuff = buffer;	  
          firstSegment = 0;
        } else {
          sbuff = buffer;
        }
        err =  XML_Parse(xmlp, sbuff, strlen(sbuff), 0);
        DB_Flush (rdfp->db) ;
        if (!err) {
          int count =   XML_GetCurrentByteIndex(xmlp); 
          err = XML_GetErrorCode(xmlp);        
          if (gVerbose) {
            printf("Parsed %i bytes (error code = %i)\n", count, err);
          }
          printf("Error parsing %s (errorcode = %i). At (%i of %i)\n", url, err, count, len);
          CLEANUP_CONNECTION;
          return RDF_ERROR;
          
        }
      }
    }
    
    if (err != 0) {
      *code = rdfp->code;
    }
    
    CLEANUP_CONNECTION 
      
      return (err == 0);    
  }  
}

char *
getFileCode(RDF_Parser rdfp) {
  if (rdfp->code) {
    return rdfp->code;
  } else {
    return 0;
  }
}


void  RDFXML_StartElementHandler (void* userData, const XML_Char *elementName, const XML_Char **attlist) ;
void RDFXML_EndElementHandler (void *userData, const XML_Char *name) ;
void  RDFXML_CharacterDataHandler(void *userData, const XML_Char *s, int len) ;

void  DMOZRDFXML_StartElementHandler (void* userData, const XML_Char *elementName, const XML_Char **attlist) ;
void DMOZRDFXML_EndElementHandler (void *userData, const XML_Char *name) ;
void  DMOZRDFXML_CharacterDataHandler(void *userData, const XML_Char *s, int len) ;


RDF_Parser
NewRDFParser (RDF_SerializationType type) {
  RDF_Parser rdfp = (RDF_Parser)getMem(sizeof(RDF_ParserStruct));

  bzero(rdfp, sizeof(RDF_ParserStruct)); 
  rdfp->classHash = NewHashTable(128);
  if (type == RSS) {
    rdfp->startElementHandler = RSS_StartElementHandler;
    rdfp->endElementHandler = RSS_EndElementHandler;
    rdfp->characterDataHandler = RSS_CharacterDataHandler;
  } else if (type == RDF_XML) {
    rdfp->startElementHandler = RDFXML_StartElementHandler;
    rdfp->endElementHandler = RDFXML_EndElementHandler;
    rdfp->characterDataHandler = RDFXML_CharacterDataHandler;
  } else if (type == DMOZ_RDF_XML) {
    rdfp->startElementHandler = DMOZRDFXML_StartElementHandler;
    rdfp->endElementHandler = DMOZRDFXML_EndElementHandler;
    rdfp->characterDataHandler = DMOZRDFXML_CharacterDataHandler;
  }
  return rdfp;
}
